#
# Copyright © 2022-2025 QPerfect. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import os
import requests
import threading
import base64
from urllib.parse import urljoin
import time

# Base class for the connection
from mimiqlink.abstractconnection import AbstractConnection

# Import the logging utils
from mimiqlink.utils import getLogger

PLANQK_API = "https://gateway.platform.planqk.de"


class JWTtoken:
    """JWT token for authentication with the PlanQK API."""

    def __init__(self, access_token, scope, token_type, expires_in):
        self.access_token = access_token
        self.scope = scope
        self.token_type = token_type
        self.expires_in = expires_in

    def __str__(self):
        return f"JWT Token:\n├── access_token: {self.access_token}\n├── scope: {self.scope}\n├── token_type: {self.token_type}\n└── expires_in: {self.expires_in}"

    def __repr__(self):
        return f"JWTtoken({self.access_token})"


class PlanqkConnection(AbstractConnection):
    """Connection to the MIMIQ services through PlanQK API.

    It handles the authentication and the requests to the PlanQK gateway.

    Examples:

    - Connect using credentials from environment variables:

    >>> conn = PlanqkConnection().connect()

    - Connect using explicit credentials:

    >>> conn = PlanqkConnection().connect("my_consumer_key", "my_consumer_secret")

    - Connect with custom URL and credentials:

    >>> conn = PlanqkConnection(url="https://custom-gateway.example.com").connect("key", "secret")
    """

    def __init__(self, url=None, consumer_key=None, consumer_secret=None):
        if url is None:
            url = PLANQK_API

        super().__init__(url)

        # PlanQK credentials
        self.consumer_key = consumer_key
        self.consumer_secret = consumer_secret

        # Token and refresher variables
        self.token = None
        self.token_lock = threading.Lock()
        self.refresher_task = None
        self.refresher_stop = False

    def get_api_url(self, *paths):
        """Get the API URL for the given paths."""
        base = urljoin(self.url + "/", "api/planqk/")
        if not paths:
            return base

        # Join all paths with the base URL
        result = base
        for path in paths:
            # Remove leading slash if present to avoid double slashes
            if path.startswith("/"):
                path = path[1:]
            result = urljoin(result + "/", path)

        return result

    @staticmethod
    def get_planqk_token(consumer_key, consumer_secret):
        """Get a new JWT token from the PlanQK API."""
        path = urljoin(PLANQK_API, "token")

        creds = base64.b64encode(f"{consumer_key}:{consumer_secret}".encode()).decode()
        headers = {"Authorization": f"Basic {creds}"}
        data = "grant_type=client_credentials"

        response = requests.post(path, headers=headers, data=data, timeout=30)

        if response.status_code != 200:
            raise ConnectionError(
                f"Failed to get PlanQK token. Server responded with {response.status_code}"
            )

        result = response.json()

        return JWTtoken(
            result["access_token"],
            result["scope"],
            result["token_type"],
            result["expires_in"],
        )

    def connect(self, consumer_key=None, consumer_secret=None):
        """Connect to the PlanQK API.

        Args:
            consumer_key: The consumer key for the PlanQK API. If None, it will use the one provided in the constructor.
            consumer_secret: The consumer secret for the PlanQK API. If None, it will use the one provided in the constructor.

        Returns:
            self: The connection object.
        """
        if self.isOpen():
            return self

        # Set the credentials if provided
        if consumer_key is not None:
            self.consumer_key = consumer_key
        if consumer_secret is not None:
            self.consumer_secret = consumer_secret

        # Check if we have credentials
        if self.consumer_key is None or self.consumer_secret is None:
            # Try to get credentials from environment variables
            self.consumer_key = self.consumer_key or os.environ.get(
                "PLANQK_CONSUMER_KEY"
            )
            self.consumer_secret = self.consumer_secret or os.environ.get(
                "PLANQK_CONSUMER_SECRET"
            )

            if self.consumer_key is None or self.consumer_secret is None:
                raise ConnectionError(
                    "No consumer key or secret provided and not found in environment variables."
                )

        # Get the initial token
        token = self.get_planqk_token(self.consumer_key, self.consumer_secret)

        with self.token_lock:
            self.token = token

        # Update the session headers
        self.__updateSessionHeaders()

        # Start the refresher
        self.__startRefresher()

        getLogger().info("Successfully connected to PlanQK API.")

        return self

    def __updateSessionHeaders(self):
        """Update the session headers with the current token."""
        with self.token_lock:
            token = self.token

        if token is None:
            raise ConnectionError("Not yet authenticated.")

        self.session.headers.update({"Authorization": f"Bearer {token.access_token}"})

    def __startRefresher(self):
        """Start the token refresher thread."""
        # If the refresher is already running, stop it
        with self.token_lock:
            if self.refresher_task is not None and self.refresher_task.is_alive():
                self.refresher_stop = True
                self.refresher_task.join()

            # Ensure that the refresher is not stopped immediately
            self.refresher_stop = False

        # Create and start the refresher
        self.refresher_task = threading.Thread(target=self.__refresherMain, daemon=True)
        self.refresher_task.start()

    def __refresherMain(self):
        """Token refresher main function.

        Will refresh the token before it expires.
        """
        while True:
            # Get the current token and its expiration time
            with self.token_lock:
                token = self.token

            if token is None:
                break

            # Sleep for 80% of the token's lifetime
            sleep_time = int(token.expires_in * 0.8)

            # Check the stop flag every second
            for _ in range(sleep_time):
                time.sleep(1)
                with self.token_lock:
                    if self.refresher_stop:
                        break

            with self.token_lock:
                if self.refresher_stop:
                    break

            # Get a new token
            try:
                getLogger().debug("Refreshing PlanQK token")
                new_token = self.get_planqk_token(
                    self.consumer_key, self.consumer_secret
                )

                with self.token_lock:
                    self.token = new_token

                self.__updateSessionHeaders()
                getLogger().debug(
                    f"Successfully refreshed PlanQK token. New expiration: {new_token.expires_in}s"
                )
            except Exception as e:
                getLogger().error(f"Failed to refresh PlanQK token: {e}")

    def close(self):
        """Close the connection to the PlanQK API."""
        getLogger().info(f"Closing connection to PlanQK API ({self.url})")

        # Ask the refresher to stop
        with self.token_lock:
            self.refresher_stop = True
            if self.refresher_task is not None and self.refresher_task.is_alive():
                getLogger().info("Shutting down token refresher")
                self.refresher_task.join()
                getLogger().info(f"Task (done) @{hex(id(self.refresher_task))}")

            # Clear the token
            self.token = None

    def isOpen(self):
        """Check if the connection is open."""
        return (
            self.refresher_task is not None
            and self.refresher_task.is_alive()
            and self.token is not None
        )

    def checkAuth(self):
        """Check if the authentication is valid."""
        with self.token_lock:
            if self.token is None:
                raise ConnectionError("Not yet authenticated to PlanQK API.")

    def __str__(self):
        result = f"PlanqkConnection:\n├── url: {self.url}\n"

        with self.token_lock:
            if self.token:
                result += f"├── token_type: {self.token.token_type}\n"
                result += f"├── expires_in: {self.token.expires_in}s\n"

        result += "└── status: " + ("open" if self.isOpen() else "closed")

        return result

    def __repr__(self) -> str:
        return self.__str__()



# FILE src/./request_info.py
#
# Copyright © 2022-2025 QPerfect. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from datetime import datetime


def format_datetime(dt_str):
    """Format a datetime string or timestamp in a more readable way."""
    if not dt_str or dt_str == "None":
        return "Not available"

    try:
        # Check if it's a timestamp (number)
        if isinstance(dt_str, (int, float)) or (
            isinstance(dt_str, str) and dt_str.isdigit()
        ):
            # Convert to integer timestamp if it's a string
            timestamp = int(dt_str) if isinstance(dt_str, str) else dt_str

            # Check if it's in milliseconds (13 digits typically)
            if timestamp > 1000000000000:  # Likely milliseconds
                timestamp = timestamp / 1000  # Convert to seconds

            # Convert timestamp to datetime
            from datetime import datetime

            dt = datetime.fromtimestamp(timestamp)
            return dt.strftime("%Y-%m-%d %H:%M:%S")

        # Try to parse ISO format if it's not a timestamp
        from datetime import datetime

        dt = datetime.fromisoformat(dt_str.replace("Z", "+00:00"))
        return dt.strftime("%Y-%m-%d %H:%M:%S")

    except (ValueError, TypeError):
        # If all conversion attempts fail, return as is
        return dt_str


class RequestInfo:
    """Class to hold and display information about a MIMIQ request."""

    STATUS_COLORS = {
        "NEW": "#3498db",  # Blue
        "RUNNING": "#f39c12",  # Orange
        "DONE": "#2ecc71",  # Green
        "ERROR": "#e74c3c",  # Red
        "CANCELED": "#95a5a6",  # Gray
    }

    def __init__(self, data):
        """
        Initialize RequestInfo with a dictionary of request data.

        Args:
            data (dict): Dictionary containing request information
        """
        self.data = data

    @property
    def id(self):
        """Get the request ID."""
        return self.data.get("_id", "Unknown")

    @property
    def name(self):
        """Get the request name."""
        return self.data.get("name", "Unknown")

    @property
    def label(self):
        """Get the request label."""
        return self.data.get("label", "Unknown")

    @property
    def status(self):
        """Get the request status."""
        return self.data.get("status", "Unknown")

    @property
    def user_email(self):
        """Get the user email."""
        return self.data.get("user", {}).get("email", "Unknown")

    @property
    def creation_date(self):
        """Get the creation date."""
        return format_datetime(self.data.get("creationDate"))

    @property
    def running_date(self):
        """Get the running date."""
        return format_datetime(self.data.get("runningDate"))

    @property
    def done_date(self):
        """Get the completion date."""
        return format_datetime(self.data.get("doneDate"))

    @property
    def num_upload_files(self):
        """Get the number of uploaded files."""
        return self.data.get("numberOfUploadedFiles", 0)

    @property
    def num_result_files(self):
        """Get the number of result files."""
        return self.data.get("numberOfResultedFiles", 0)

    def get(self, key, default):
        """Get a specific attribute from the request data."""
        return self.data.get(key, default)

    def __repr__(self):
        """Return a single line string representation of the request info."""
        result = (
            f"Request {self.id} | Name: {self.name} | Label: {self.label} | "
            f"Status: {self.status} | User: {self.user_email} | "
            f"Created: {self.creation_date} | Running: {self.running_date} | "
            f"Completed: {self.done_date}"
        )

        if self.num_upload_files > 0 or self.num_result_files > 0:
            result += (
                f" | Files: {self.num_upload_files}/{self.num_result_files} (up/res)"
            )

        return result

    def _repr_html_(self):
        """
        Return an HTML representation of the request info.
        This will be automatically used by Jupyter notebooks.
        """
        status_color = self.STATUS_COLORS.get(self.status, "#95a5a6")

        html = f"""
        <div style="margin: 10px; padding: 10px; border: 1px solid #ddd; border-radius: 5px; max-width: 600px;">
            <h3 style="margin-top: 0; color: #333;">Request {self.id}</h3>
            <table style="width: 100%; border-collapse: collapse;">
                <tr>
                    <td style="padding: 4px 10px 4px 0; font-weight: bold;">Name:</td>
                    <td>{self.name}</td>
                </tr>
                <tr>
                    <td style="padding: 4px 10px 4px 0; font-weight: bold;">Label:</td>
                    <td>{self.label}</td>
                </tr>
                <tr>
                    <td style="padding: 4px 10px 4px 0; font-weight: bold;">Status:</td>
                    <td><span style="background-color: {status_color}; color: white; padding: 2px 6px; border-radius: 3px;">{self.status}</span></td>
                </tr>
                <tr>
                    <td style="padding: 4px 10px 4px 0; font-weight: bold;">User Email:</td>
                    <td>{self.user_email}</td>
                </tr>
                <tr>
                    <td style="padding: 4px 10px 4px 0; font-weight: bold;">Created:</td>
                    <td>{self.creation_date}</td>
                </tr>
                <tr>
                    <td style="padding: 4px 10px 4px 0; font-weight: bold;">Running:</td>
                    <td>{self.running_date}</td>
                </tr>
                <tr>
                    <td style="padding: 4px 10px 4px 0; font-weight: bold;">Completed:</td>
                    <td>{self.done_date}</td>
                </tr>
        """

        if self.num_upload_files > 0:
            html += f"""
                <tr>
                    <td style="padding: 4px 10px 4px 0; font-weight: bold;">Uploaded Files:</td>
                    <td>{self.num_upload_files}</td>
                </tr>
            """

        if self.num_result_files > 0:
            html += f"""
                <tr>
                    <td style="padding: 4px 10px 4px 0; font-weight: bold;">Result Files:</td>
                    <td>{self.num_result_files}</td>
                </tr>
            """

        html += """
            </table>
        </div>
        """
        return html


class RequestInfoList:
    """Class to hold and display a list of MIMIQ requests."""

    def __init__(self, data_list):
        """
        Initialize RequestInfoList with a list of request data dictionaries.

        Args:
            data_list (list): List of dictionaries containing request information
        """
        self.requests = [RequestInfo(data) for data in data_list]

    def __len__(self):
        """Return the number of requests in the list."""
        return len(self.requests)

    def __getitem__(self, idx):
        """Get the request at the specified index."""
        return self.requests[idx]

    def __iter__(self):
        """Return an iterator over the requests."""
        return iter(self.requests)

    @property
    def status_counts(self):
        """Get a dictionary with counts of each status."""
        counts = {}
        for req in self.requests:
            status = req.status
            counts[status] = counts.get(status, 0) + 1
        return counts

    def __repr__(self):
        """Return a string representation of the request list."""
        if not self.requests:
            return "No requests available"

        # Create summary line
        running_count = self.status_counts.get("RUNNING", 0)
        new_count = self.status_counts.get("NEW", 0)
        done_count = self.status_counts.get("DONE", 0)
        error_count = self.status_counts.get("ERROR", 0)
        canceled_count = self.status_counts.get("CANCELED", 0)

        summary = f"Total: {len(self.requests)} requests - "
        status_parts = []
        if new_count > 0:
            status_parts.append(f"{new_count} NEW")
        if running_count > 0:
            status_parts.append(f"{running_count} RUNNING")
        if done_count > 0:
            status_parts.append(f"{done_count} DONE")
        if error_count > 0:
            status_parts.append(f"{error_count} ERROR")
        if canceled_count > 0:
            status_parts.append(f"{canceled_count} CANCELED")

        summary += ", ".join(status_parts)

        # Format header line with wider ID column (24 characters + some padding)
        header = "ID                          LABEL                STATUS"

        # Create the result list with summary and header
        result = [summary, header, "-" * len(header)]

        # Add a single line for each request
        for req in self.requests:
            # Format each field with fixed width (ID width increased to 26)
            req_id = f"{req.id:<26}"
            label = f"{req.label[:18]:<18}" + ("..." if len(req.label) > 18 else "  ")
            status = f"{req.status:<8}"

            line = f"{req_id}{label}{status}"
            result.append(line)

        return "\n".join(result)

    def _repr_html_(self):
        """
        Return an HTML representation of the request list as a table.
        This will be automatically used by Jupyter notebooks.
        """
        if not self.requests:
            return "<p>No requests available</p>"

        # Summary section
        status_summary = " | ".join(
            f"{status}: <b>{count}</b>" for status, count in self.status_counts.items()
        )

        html = f"""
        <div style="margin: 10px; padding: 10px; border: 1px solid #ddd; border-radius: 5px; max-width: 100%;">
            <h2 style="margin-top: 0; color: #333;">Request Summary</h2>
            <p>Total: <b>{len(self.requests)}</b> requests ({status_summary})</p>
            <div style="max-height: 800px; overflow-y: auto; overflow-x: auto;">
                <table style="width: 100%; border-collapse: collapse; font-size: 14px;">
                    <thead>
                        <tr style="background-color: #f8f9fa; border-bottom: 2px solid #ddd;">
                            <th style="text-align: left; padding: 8px; white-space: nowrap;">ID</th>
                            <th style="text-align: left; padding: 8px; white-space: nowrap;">Name</th>
                            <th style="text-align: left; padding: 8px; white-space: nowrap;">Label</th>
                            <th style="text-align: center; padding: 8px; white-space: nowrap;">Status</th>
                            <th style="text-align: left; padding: 8px; white-space: nowrap;">User Email</th>
                            <th style="text-align: left; padding: 8px; white-space: nowrap;">Created</th>
                            <th style="text-align: left; padding: 8px; white-space: nowrap;">Running</th>
                            <th style="text-align: left; padding: 8px; white-space: nowrap;">Completed</th>
                            <th style="text-align: center; padding: 8px; white-space: nowrap;">Upload Files</th>
                            <th style="text-align: center; padding: 8px; white-space: nowrap;">Result Files</th>
                        </tr>
                    </thead>
                    <tbody>
        """

        # Add each request as a row in the table
        for i, req in enumerate(self.requests):
            status_color = RequestInfo.STATUS_COLORS.get(req.status, "#95a5a6")
            row_style = "background-color: #f8f8f8;" if i % 2 == 0 else ""

            html += f"""
                        <tr style="{row_style}">
                            <td style="padding: 8px; border-bottom: 1px solid #ddd;">{req.id}</td>
                            <td style="padding: 8px; border-bottom: 1px solid #ddd;">{req.name}</td>
                            <td style="padding: 8px; border-bottom: 1px solid #ddd;">{req.label}</td>
                            <td style="padding: 8px; border-bottom: 1px solid #ddd; text-align: center;"><span style="background-color: {status_color}; color: white; padding: 2px 6px; border-radius: 3px;">{req.status}</span></td>
                            <td style="padding: 8px; border-bottom: 1px solid #ddd;">{req.user_email}</td>
                            <td style="padding: 8px; border-bottom: 1px solid #ddd;">{req.creation_date}</td>
                            <td style="padding: 8px; border-bottom: 1px solid #ddd;">{req.running_date}</td>
                            <td style="padding: 8px; border-bottom: 1px solid #ddd;">{req.done_date}</td>
                            <td style="padding: 8px; border-bottom: 1px solid #ddd; text-align: center;">{req.num_upload_files}</td>
                            <td style="padding: 8px; border-bottom: 1px solid #ddd; text-align: center;">{req.num_result_files}</td>
                        </tr>
            """

        html += """
                    </tbody>
                </table>
            </div>
        </div>
        """

        return html



#
# Copyright © 2022-2025 QPerfect. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from abc import ABC, abstractmethod
import os
import os.path
import io
import re
import requests
from requests.adapters import HTTPAdapter

# Import the logging utils
from mimiqlink.utils import getLogger

from mimiqlink.infos import RequestInfo, RequestInfoList


class TimeoutHTTPAdapter(HTTPAdapter):
    def __init__(self, *args, **kwargs):
        if "timeout" in kwargs:
            self.timeout = kwargs["timeout"]
            del kwargs["timeout"]
        super().__init__(*args, **kwargs)

    def send(self, request, **kwargs):
        timeout = kwargs.get("timeout")
        if timeout is None and hasattr(self, "timeout"):
            kwargs["timeout"] = self.timeout
        # HACK: fix here. used to put timeout=0 (and hence timeout=None) in upload (otherwise we have a timeout)
        if timeout == 0:
            kwargs["timeout"] = None
        return super().send(request, **kwargs)


class ConnectionError(Exception):
    """Exception raised for errors in the connection to the remote server."""

    pass


class AbstractConnection(ABC):
    """Base class for connection to MIMIQ and other remote services.

    This abstract class defines the interface for all connection types.
    """

    def __init__(self, url=None):
        self.url = url

        # session for doing requests
        self.session = requests.Session()
        self.session.mount("http://", TimeoutHTTPAdapter(timeout=(1, None)))
        self.session.mount("https://", TimeoutHTTPAdapter(timeout=(1, None)))

    @abstractmethod
    def connect(self, *args, **kwargs):
        """Establish a connection to the remote service."""
        pass

    @abstractmethod
    def close(self):
        """Close the connection to the remote service."""
        pass

    @abstractmethod
    def isOpen(self):
        """Check if the connection is open."""
        pass

    @abstractmethod
    def checkAuth(self):
        """Check if the authentication is valid."""
        pass

    @abstractmethod
    def get_api_url(self, *paths):
        """Get the API URL for the given paths.

        This method should join the base URL with the API prefix and the given paths.
        """
        pass

    def request(self, emulatortype, name, label, timeout, uploads):
        """Request an execution to the remote server."""
        self.checkAuth()

        endpoint = "request"

        data = [
            ("name", (None, name)),
            ("label", (None, label)),
            ("emulatorType", (None, emulatortype)),
            ("timeout", (None, timeout)),
        ]

        for file in uploads:
            if isinstance(file, io.IOBase) and not file.closed:
                data.append(("uploads", (os.path.basename(file.name), file)))
            else:
                data.append(("uploads", (os.path.basename(file), open(file, "rb"))))

        response = self.session.post(self.get_api_url(endpoint), files=data, timeout=0)

        if response.status_code != 200:
            raise ConnectionError(
                f"File upload failed with status code {response.status_code}"
            )

        return response.json()["executionRequestId"]

    def requestInfo(self, request):
        """Retrieve the execution details for a given request."""
        self.checkAuth()

        endpoint = f"request/{request}"

        response = self.session.get(self.get_api_url(endpoint))

        if response.status_code != 200:
            raise ConnectionError(
                f"Failed to retrieve execution details for {request}. Server responded with {response.status_code}"
            )

        return RequestInfo(response.json())

    def requests(self, **kwargs):
        """Retrieve the list of requests from the server."""
        self.checkAuth()

        query = ""
        for key, value in kwargs.items():
            if query == "":
                query += f"?{key}={value}"
            else:
                query += f"&{key}={value}"

        endpoint = "request"
        response = self.session.get(self.get_api_url(endpoint) + query)
        if response.status_code != 200:
            raise ConnectionError(
                f"Failed to retrieve the list of requests. Server responded with {response.status_code}"
            )

        return RequestInfoList(response.json()["executions"]["docs"])

    def isJobDone(self, request):
        """Check if the job is done."""
        infos = self.requestInfo(request)
        status = infos.status
        return status == "DONE" or status == "ERROR"

    def isJobFailed(self, request):
        """Check if the job failed."""
        infos = self.requestInfo(request)
        status = infos.status
        return status == "ERROR"

    def isJobStarted(self, request):
        """Check if the job is started."""
        infos = self.requestInfo(request)
        status = infos.status
        return status != "NEW"

    def isJobCanceled(self, request):
        """Check if the job is canceled."""
        infos = self.requestInfo(request)
        status = infos.status
        return status == "CANCELED"

    def stopexecution(self, request):
        """Stop the execution of a given request."""
        self.checkAuth()

        endpoint = f"stop-execution/{request}"
        response = self.session.post(self.get_api_url(endpoint))

        if response.status_code != 200:
            raise ConnectionError(
                f"Failed to stop the execution {request}. Server responded with {response.status_code}."
            )

        return None

    def deleteFiles(self, request):
        """Delete the files for a given request."""
        self.checkAuth()

        endpoint = f"delete-files/{request}"
        response = self.session.post(self.get_api_url(endpoint))

        if response.status_code != 200:
            raise ConnectionError(
                f"Failed to delete the files for {request}. Server responded with {response.status_code}."
            )

        return None

    def downloadFile(self, request, index, filetype, destdir):
        """Download a specific file from the server."""
        self.checkAuth()

        endpoint = f"files/{request}/{index}"
        url = self.get_api_url(endpoint) + f"?source={filetype}"

        response = self.session.get(url)

        if response.status_code >= 300:
            raise ConnectionError(
                f"Failed to retrieve {filetype} files for {request}. Server responded with {response.status_code}"
            )

        filename = re.findall(
            'filename="(.+)"', response.headers.get("Content-Disposition")
        )[0]

        # Should never happen, but just in case.
        # If it does, we can't do anything about it here. We need to patch the server
        if not filename:
            raise ConnectionError(
                "Something went wrong. Server is missing the filename"
            )

        # Create directory if it doesn't exist
        if not os.path.exists(destdir):
            os.makedirs(destdir)

        # at this point we should have a valid filename and a valid directory
        # so we write everything to the file
        with open(os.path.join(destdir, filename), "wb") as f:
            f.write(response.content)

        return filename

    def downloadFiles(self, request, source, destdir=None):
        """Download files for a given request."""
        self.checkAuth()

        if destdir is None:
            destdir = os.path.join("./", request)

        # Create the directory if it doesn't exist
        if not os.path.exists(destdir):
            os.makedirs(destdir)

        infos = self.requestInfo(request)

        if source == "uploads":
            sourcename = "numberOfUploadedFiles"
        else:
            sourcename = "numberOfResultedFiles"

        nf = infos.get(sourcename, 0)

        names = []
        for idx in range(nf):
            name = self.downloadFile(request, idx, source, destdir)
            names.append(name)

        return names

    def downloadJobFiles(self, request, **kwargs):
        """Download the input files for a given request."""
        return self.downloadFiles(request, "uploads", **kwargs)

    def downloadResults(self, request, **kwargs):
        """Download the result files for a given request."""
        return self.downloadFiles(request, "results", **kwargs)



#
# Copyright © 2022-2024 University of Strasbourg. All Rights Reserved.
# Copyright © 2022-2025 QPerfect. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from functools import partial
from http.server import HTTPServer
import json
import threading
import webbrowser
from urllib.parse import urljoin
import time
from mimiqlink.abstractconnection import AbstractConnection

# Import the connection handler
from mimiqlink.handler import AuthenticationHandler

# Import the logging utils
from mimiqlink.utils import getLogger

QPERFECT_CLOUD = "https://mimiq.qperfect.io"
QPERFECT_DEV = "https://mimiqfast.qperfect.io"


class MimiqConnection(AbstractConnection):
    """Connection to the MIMIQ remote services.

    It handles the authentication and the requests.
    """

    def __init__(self, url=None):
        if url is None:
            url = QPERFECT_CLOUD

        super().__init__(url)

        # refresher related variables
        self.refresher_lock = threading.Lock()
        self.refresher_task = None
        self.refresher_interval = 15 * 60
        self.refresher_stop = False

        # tokens
        self.access_token = None
        self.refresh_token = None

        # user limits
        self.user_limits = None

    def get_api_url(self, *paths):
        """Get the API URL for the given paths."""
        base = urljoin(self.url + "/", "api")
        if not paths:
            return base

        # Join all paths with the base URL
        result = base
        for path in paths:
            # Remove leading slash if present to avoid double slashes
            if path.startswith("/"):
                path = path[1:]
            result = urljoin(result + "/", path)

        return result

    def _weblogin(self, data):
        """Authenticate to the remote server with the given credentials. But return the response."""
        endpoint = "sign-in"

        # ask for access tokens
        response = self.session.post(
            self.get_api_url(endpoint), json=data, headers={"Connection": "close"}
        )

        if response.status_code != 200:
            reason = response.json().get("message", "Unknown error")
            raise ConnectionError(
                f"Authentication failed with status code {response.status_code} and reason: {reason}"
            )

        tokens = response.json()

        # set the access tokens
        with self.refresher_lock:
            self.access_token = tokens["token"]
            self.refresh_token = tokens["refreshToken"]

        self.__updateSessionHeaders()
        self.__updateUserLimits()
        self.__startRefresher()

        getLogger().info("Authentication successful.")

        return response

    def connectUser(self, email, password):
        """Authenticate to the remote server with the given credentials (email and password)."""

        if self.isOpen():
            self.close()

        self._weblogin({"email": email, "password": password})

        return self

    def connectToken(self, token):
        """Authenticate to the remote server with the given refresh token"""

        if self.isOpen():
            self.close()

        # set the refresh token
        with self.refresher_lock:
            self.refresh_token = token

        # refresh
        status = self.refresh()

        if not status:
            raise ConnectionError("Authentication failed.")

        getLogger().info("Authentication successful.")
        self.__updateSessionHeaders()
        self.__updateUserLimits()
        self.__startRefresher()
        return self

    def connectWeb(self):
        """Authenticate to the remote services by taking credentials from a locally shown login page"""

        if self.isOpen():
            self.close()

        preferred_port = 1444

        h = partial(AuthenticationHandler, lambda data: self._weblogin(data))

        try:
            # Attempt to create the server with the fixed port
            httpd = HTTPServer(("localhost", preferred_port), h)
            port = preferred_port
        except OSError:
            # If the fixed port is in use, use a random available port
            httpd = HTTPServer(("localhost", 0), h)
            port = httpd.server_port

        getLogger().info(
            f"Listening on: 127.0.0.1:{port}, thread id: {threading.get_ident()}"
        )
        getLogger().info(f"Please login in your browser at http://localhost:{port}")

        webbrowser.open(f"http://localhost:{port}", new=2)

        with httpd:
            while self.access_token is None:
                httpd.handle_request()

        if not self.access_token:
            raise ConnectionError(
                "Authentication failed. Unable to obtain access token."
            )

        return self

    def connect(self, *args, **kwargs):
        """Connect to the remote server.

        If no arguments are provided, the connection will be established using the web login.
        If one argument is provided, it will be used as the token.
        If two arguments are provided, they will be used as the email and password.

        Examples:

        - Connect using the web login:

        >>> conn = MimiqConnection()
        >>> conn.connect()

        or simply:

        >>> conn = MimiqConnection().connect()

        - Connect using a token:

        >>> conn = MimiqConnection()
        >>> conn.connect("mytoken")

        or simply:

        >>> conn = MimiqConnection().connect("mytoken")

        - Connect using a token and a user:

        >>> conn = MimiqConnection()
        >>> conn.connect("john.doe@example.com", "password")
        """
        if self.isOpen():
            return self

        if len(args) == 0:
            return self.connectWeb()

        if len(args) == 1:
            return self.connectToken(*args)

        if len(args) == 2:
            return self.connectUser(*args)

        raise ConnectionError(
            "Invalid number of arguments. Expected 0, 1 (token) or 2 (username, password)."
        )

    def __startRefresher(self):
        """Start the refresher thread."""

        # if the refresher is alreeady running stop it
        with self.refresher_lock:
            if self.refresher_task is not None and self.refresher_task.is_alive():
                self.refresher_stop = True
                self.refresher_task.join()

        # ensure that the refresher is not stopped immediately
        with self.refresher_lock:
            self.refresher_stop = False

        # create and start the refresher
        self.refresher_task = threading.Thread(target=self.__refresherMain, daemon=True)
        self.refresher_task.start()

    def __refresherMain(self):
        """Refresher function
        Will refresh the access token with the refresh token at every
        configured interval.
        """
        while True:
            # check the stop flag every second
            for i in range(self.refresher_interval):
                time.sleep(1)
                with self.refresher_lock:
                    if self.refresher_stop:
                        break

            with self.refresher_lock:
                if self.refresher_stop:
                    break

            self.refresh()

    def refresh(self):
        """Refresh the access token using the refresh token."""
        endpoint = "/access-token"

        # prepare the request
        with self.refresher_lock:
            data = {"refreshToken": self.refresh_token}

        # ask for a new access token
        response = self.session.post(
            self.get_api_url(endpoint), json=data, headers={"Connection": "close"}
        )

        # check if the response is valid
        if response.status_code != 200:
            raise ConnectionError(
                f"Failed to refresh the access token. Server responded with {response.status_code}"
            )

        tokens = response.json()

        # write the new tokens
        with self.refresher_lock:
            self.access_token = tokens["token"]
            self.refresh_token = tokens["refreshToken"]

        self.__updateSessionHeaders()
        self.__updateUserLimits()
        return True

    def __updateSessionHeaders(self):
        # fetch the access token
        with self.refresher_lock:
            token = self.access_token

        # build the headers
        self.session.headers.update({"Authorization": f"Bearer {token}"})

    def checkAuth(self):
        with self.refresher_lock:
            if self.access_token is None:
                raise ConnectionError("Not yet authenticated.")

    def savetoken(self, filepath="qperfect.json"):
        """Save the current token to a file."""
        if self.refresh_token is None:
            self.connect()
        with open(filepath, "w") as f:
            json.dump({"token": self.refresh_token, "url": self.url}, f)

    def loadtoken(self, filepath="qperfect.json"):
        """Load and connect using a token from a file."""
        # Attempt to read the token file
        try:
            with open(filepath, "r") as f:
                data = json.load(f)
                saved_url = data.get("url", "")
                token = data.get("token")

            # Check if the current URL matches the saved URL in the token file
            if self.url != saved_url:
                raise ValueError(
                    f"The URL in the token file ({saved_url}) does not match the current URL ({self.url}). Authentication failed."
                )

        except Exception as e:
            # Log error and re-raise as ConnectionError
            getLogger().error(f"Error reading token file: {e}")
            raise ConnectionError("Failed to read token file.") from e

        # Establish a new connection using the token
        try:
            self.connectToken(token)
            return self
        except ConnectionError as e:
            # Log error and re-raise
            getLogger().error(f"Authentication failed: {e}")
            raise ConnectionError(
                "Authentication failed. Unable to connect using the stored token."
            ) from e

    def close(self):
        """Close the connection."""
        getLogger().info(f"Closing connection to {self.url}")
        # ask the refresher to stop
        with self.refresher_lock:
            self.refresher_stop = True
            if self.refresher_task is not None and self.refresher_task.is_alive():
                getLogger().info("Shutting down token refresher")
                self.refresher_task.join()
                getLogger().info(f"Task (done) @{hex(id(self.refresher_task))}")

        # clean the tokens
        self.access_token = None
        self.refresh_token = None

    def isOpen(self):
        """Check if the connection is open."""
        return (
            self.refresher_task is not None
            and self.refresher_task.is_alive()
            and self.access_token is not None
        )

    def checkUserLimits(self, limits=None):
        """Check if the user limits are respected."""
        if limits is None:
            limits = self.user_limits

        if limits is None:
            return

        if limits.get("enabledExecutionTime"):
            used_time = limits.get("usedExecutionTime")
            max_time = limits.get("maxExecutionTime")
            if used_time > max_time:
                getLogger().warning(
                    f"You have exceeded your computing time limit of {max_time} minutes"
                )

        if limits.get("enabledMaxExecutions"):
            used_exec = limits.get("usedExecutions")
            max_exec = limits.get("maxExecutions")
            if used_exec is not None and max_exec is not None:
                if used_exec > max_exec:
                    getLogger().warning(
                        f"You have exceeded your number of executions limit of {max_exec}"
                    )

    def __updateUserLimits(self):
        """Fetch user limits from the server."""
        self.checkAuth()

        endpoint = "users/limits"

        response = self.session.get(self.get_api_url(endpoint))

        if response.status_code != 200:
            raise ConnectionError(
                f"Failed to retrieve user limits. Server responded with {response.status_code}"
            )

        self.user_limits = response.json()
        self.checkUserLimits()

    def __str__(self):
        limits = self.user_limits
        result = f"MimiqConnection:\n├── url: {self.url}\n"

        if limits is not None:
            if limits.get("enabledExecutionTime"):
                used_time = limits.get("usedExecutionTime")
                max_time = limits.get("maxExecutionTime")
                if used_time is not None and max_time is not None:
                    result += f"├── Computing time: {round(used_time / 60)}/{round(max_time / 60)} minutes\n"
            if limits.get("enabledMaxExecutions"):
                used_exec = limits.get("usedExecutions")
                max_exec = limits.get("maxExecutions")
                if used_exec is not None and max_exec is not None:
                    result += f"├── Executions: {used_exec}/{max_exec}\n"
            if limits.get("enabledMaxTimeout"):
                max_timeout = limits.get("maxTimeout")
                if max_timeout is not None:
                    max_timeout = round(max_timeout)
                    result += f"├── Max time limit per request: {max_timeout} minutes\n"
                    result += f"├── Default time limit is equal to max time limit: {max_timeout} minutes\n"
            else:
                result += "├── Max time limit is: 30 minutes\n"
                result += "├── Default time limit is: 30 minutes\n"

        result += "└── status: " + ("open" if self.isOpen() else "closed")

        return result

    def __repr__(self) -> str:
        return self.__str__()



#
# Copyright © 2022-2025 QPerfect. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import logging
import sys

def getLogger():
    logger = logging.getLogger("mimiqlink")

    # Only add handler if it's not already added
    if not logger.hasHandlers():
        handler = logging.StreamHandler(sys.stdout)
        handler.setLevel(logging.INFO)
        formatter = logging.Formatter("[%(levelname)s] %(message)s")
        handler.setFormatter(formatter)
        logger.addHandler(handler)

    logger.setLevel(logging.INFO)
    return logger




#
# Copyright © 2022-2024 University of Strasbourg. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from mimiqlink.abstractconnection import (
    AbstractConnection,
)

from mimiqlink.mimiqconnection import (
    MimiqConnection,
    QPERFECT_CLOUD,
    QPERFECT_DEV,
)

from mimiqlink.planqkconnection import (
    PlanqkConnection,
    PLANQK_API,
)

__all__ = [
    "AbstractConnection",
    "MimiqConnection",
    "PlanqkConnection",
    "QPERFECT_CLOUD",
    "QPERFECT_DEV",
    "PLANQK_API",
]



#
# Copyright © 2022-2024 University of Strasbourg. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from http.server import BaseHTTPRequestHandler
import json
import mimetypes
import os


class AuthenticationHandler(BaseHTTPRequestHandler):
    def __init__(self, authenticate_function, *args, **kwargs):
        self.authenticate_function = authenticate_function
        self._status_code = None
        # BaseHTTPRequestHandler calls do_GET **inside** __init__ !!!
        # So we have to call super().__init__ after setting attributes.
        super().__init__(*args, **kwargs)

    def send_response(self, code, message=None):
        self._status_code = code
        super().send_response(code, message)

    def do_GET(self):
        # try to serve local files in the /pubic folder
        try:
            if self.path == "/":
                self.path = "/index.html"

            public_dir = os.path.join(os.path.dirname(__file__), "public")
            filepath = os.path.join(public_dir, self.path.lstrip("/"))

            with open(filepath, "rb") as file:
                content = file.read()
                mimetype, _ = mimetypes.guess_type(filepath)
                if not mimetype:
                    mimetype = ""
                self.send_response(200)
                self.send_header("Content-type", mimetype)
                self.end_headers()
                self.wfile.write(content)
        except FileNotFoundError:
            self.send_error(404, "File Not Found: %s" % self.path)
        except Exception as e:
            self.send_error(500, "Internal Server Error: %s" % str(e))

    def do_POST(self):
        if self.path == "/api/login":
            content_length = int(self.headers["Content-Length"])
            body = self.rfile.read(content_length)
            try:
                # get the remote response from the authentication function and
                # then forward it to the local page in order to show proper
                # error messages.
                data = json.loads(body)
                response = self.authenticate_function(data)
                self.send_response(response.status_code)
                # FIX: should be the same content-type as the response?
                self.send_header("Content-type", "application/json")
                self.end_headers()
                self.wfile.write(response.content)
            except json.JSONDecodeError:
                self.send_error(400, "Bad Request: Unable to parse JSON")
            except Exception as e:
                self.send_error(500, "Internal Server Error: %s" % str(e))
        else:
            self.send_error(404, "Not Found: %s" % self.path)

    def log_message(self, format, *args):
        # Log only if the status code is not 200
        if self._status_code != 200:
            super().log_message(format, *args)
